__all__ = ["binning", "debugging", "experiment", "experimentdata", "results", "statistics", "util", "version"]

from version import __version__, version

print 'ExpAn core init: {}'.format(version())
